/*
 * Support for initialising programs in new tasks and threads.
 *
 * Copyright (C) 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/sys/err.h>
#include <l4/sys/scheduler.h>

#include <exec/process.h>
#include <ipc/cap_alloc.h>
#include <ipc/map.h>
#include <mem/memory_utils.h>



/* Obtain a flexpage defining the UTCB area location and size in a new task. */

static l4_fpage_t get_utcb_fpage()
{
  /* UTCB location and size. */

  int utcb_log2size = page_order(Default_max_threads * L4_UTCB_OFFSET);

  /* Round up to at least one page. */

  if (utcb_log2size < L4_PAGESHIFT)
    utcb_log2size = L4_PAGESHIFT;

  return l4_fpage(Utcb_area_start, utcb_log2size, 0);
}



/* Initialise a new process, this being an abstraction for a new task with some
   threads. */

Process::Process()
{
  /* Obtain UTCB area details for the task. */

  l4_fpage_t utcb_fpage = get_utcb_fpage();

  _utcb_start = l4_fpage_memaddr(utcb_fpage);

  /* Populate the common initial environment for the threads. */

  _env.factory          = L4_BASE_FACTORY_CAP;
  _env.main_thread      = L4_BASE_THREAD_CAP;
  _env.log              = L4_BASE_LOG_CAP;
  _env.scheduler        = L4_BASE_SCHEDULER_CAP;
  _env.rm               = L4_EXEC_RM_CAP;
  _env.mem_alloc        = L4_EXEC_MA_CAP;
  _env.first_free_cap   = L4_EXEC_FIRST_FREE_CAP_INDEX;
  _env.utcb_area        = utcb_fpage;
  _env.first_free_utcb  = l4_fpage_memaddr(utcb_fpage) + L4_UTCB_OFFSET;

  /* Populate auxiliary information. */

  _aux.kip_ds           = L4_EXEC_KIP_CAP;
  _aux.dbg_lvl          = 0;
  _aux.ldr_flags        = 0;
}

/* Task and thread initialisation. */

long Process::create_task()
{
  _task = ipc_cap_alloc();

  if (l4_is_invalid_cap(_task))
    return -L4_ENOMEM;

  return l4_error(l4_factory_create_task(l4re_env()->factory, _task, _env.utcb_area));
}

long Process::create_thread(l4_cap_idx_t *thread)
{
  *thread = ipc_cap_alloc();

  if (l4_is_invalid_cap(*thread))
    return -L4_ENOMEM;

  return l4_error(l4_factory_create_thread(l4re_env()->factory, *thread));
}

/* Configure the task environment. */

long Process::configure(l4_cap_idx_t server)
{
  long err = create_task();

  if (err)
    return err;

  /* Map the KIP into the task. */

  l4_addr_t kip_start = (l4_addr_t) l4re_kip();

  err = l4_error(l4_task_map(_task, L4RE_THIS_TASK_CAP,
                             l4_fpage(kip_start, L4_PAGESHIFT, L4_FPAGE_RX),
                             kip_start));

  if (err)
    return err;

  /* Define capability mappings for the new task. */

  struct ipc_mapped_cap mapped_caps[] = {
    {L4_BASE_TASK_CAP,  _task,                  L4_CAP_FPAGE_RWS},
    {L4_EXEC_PAGER_CAP, server,                 L4_CAP_FPAGE_RWS},
    {_env.rm,           server,                 L4_CAP_FPAGE_RWS},
    {_env.factory,      l4re_env()->factory,    L4_CAP_FPAGE_RWS},
    {_env.log,          l4re_env()->log,        L4_CAP_FPAGE_RWS},
    {_env.scheduler,    l4re_env()->scheduler,  L4_CAP_FPAGE_RWS},
    {_env.mem_alloc,    l4re_env()->mem_alloc,  L4_CAP_FPAGE_RWS},
    {0,                 L4_INVALID_CAP,         0},
    };

  return ipc_map_capabilities(_task, mapped_caps);
}

/* Create, initialise and start a thread. */

long Process::thread_start(l4_addr_t program_start, Stack &st)
{
  l4_cap_idx_t thread;
  long err;

  err = create_thread(&thread);

  if (err)
    return err;

  /* Initialise the thread with pager, UTCB and task details. */

  l4_thread_control_start();
  l4_thread_control_pager(L4_EXEC_PAGER_CAP);
  l4_thread_control_exc_handler(L4_EXEC_PAGER_CAP);
  l4_thread_control_bind((l4_utcb_t *) _utcb_start, _task);

  err = l4_error(l4_thread_control_commit(thread));

  if (err)
  {
    ipc_cap_free(thread);
    return err;
  }

  /* Map the thread capability to the task. */

  ipc_map_capability(_task, (struct ipc_mapped_cap) {_env.main_thread, thread, L4_CAP_FPAGE_RWS});

  /* Populate the initial environment in the thread. */

  st.set_l4re_aux(&_aux);
  st.set_l4re_env(&_env);

  /* Set the start details. */

  err = l4_error(l4_thread_ex_regs(thread, program_start, st.start_address(), 0));

  if (err)
    return err;

  /* Select a new address for the next thread. */

  _utcb_start += L4_UTCB_OFFSET;

  /* Start the thread. */

  l4_sched_param_t sp = l4_sched_param(L4RE_MAIN_THREAD_PRIO);

  return l4_error(l4_scheduler_run_thread(l4re_env()->scheduler, thread, &sp));
}

/* vim: tabstop=2 expandtab shiftwidth=2
*/
